/* Verzeichnis rekursiv einlesen und alle Dateien nach Gre sortieren        */

/*----------------------------------------------------------------------------*/
/* System-Includes                                                            */
/*----------------------------------------------------------------------------*/

	#include <proto/exec.h>
	#include <proto/dos.h>
	#include <stdlib.h>
	#include <string.h>

/*----------------------------------------------------------------------------*/
/* Typdefinitionen                                                            */
/*----------------------------------------------------------------------------*/

	struct verzeichnis {
		struct MinNode node;
		char name[1];
		};

	struct datei {
		struct MinNode node;
		struct verzeichnis *mutter;
		ULONG groesse;
		char name[1];
		};

/*----------------------------------------------------------------------------*/
/* Globale Variablen                                                          */
/*----------------------------------------------------------------------------*/

	struct MinList verzeichnisse;
	struct MinList dateien;

	BOOL verbose = FALSE;

/*----------------------------------------------------------------------------*/
/* Prfen ob Ctrl-C gedrckt wurde                                            */
/*----------------------------------------------------------------------------*/

	BOOL abbruch_pruefen (void)
	{
	static BOOL abbruch = FALSE;

	if (!abbruch)
		if (SetSignal(0,0) & SIGBREAKF_CTRL_C)
			{
			Printf ("*** Abbruch\n");
			abbruch = TRUE;
			}

	return (abbruch);
	}

/*----------------------------------------------------------------------------*/
/* Ein neues Verzeichnis ans Ende der Liste anhngen                          */
/*----------------------------------------------------------------------------*/

	void neues_verzeichnis (struct verzeichnis *mutter,char *name)

	{
	struct verzeichnis *neu;
	long len = (mutter ? strlen(mutter->name) + 1 : 0) + strlen(name);

	if (neu = malloc (sizeof(struct verzeichnis) + len))
		{
		if (mutter)
			{
			strcpy (neu->name,mutter->name);
			AddPart (neu->name,name,len+1);
			}
		else
			strcpy (neu->name,name);

		if (verbose) Printf ("neues Verzeichnis: %s\n",neu->name);

		AddTail ((struct List *)&verzeichnisse,(struct Node *)neu);
		}
	}

/*----------------------------------------------------------------------------*/
/* Eine neue Datei in die Liste einsortieren                                  */
/*----------------------------------------------------------------------------*/

	void neue_datei (struct verzeichnis *mutter,char *name,ULONG groesse)

	{
	struct datei *neu;
	struct datei *vor,*naechste;

	if (neu = malloc (sizeof(struct datei) + strlen(name)))
		{
		strcpy (neu->name,name);
		neu->mutter = mutter;
		neu->groesse = groesse;

		if (verbose) Printf ("neue Datei: %s\n",neu->name);

		for (vor = (struct datei *)dateien.mlh_Head; naechste = (struct datei *)vor->node.mln_Succ; vor = naechste)
			{
			if (vor->groesse > neu->groesse)
				break;
			}

		if (naechste)
			Insert ((struct List *)&dateien,(struct Node *)neu,(struct Node *)vor->node.mln_Pred);
		else
			AddTail ((struct List *)&dateien,(struct Node *)neu);
		}
	}

/*----------------------------------------------------------------------------*/
/* Ein Verzeichnis und dessen Unterverzeichnisse lesen                        */
/*----------------------------------------------------------------------------*/

	void lies_verzeichnis (struct verzeichnis *mutter)

	{
	BPTR lock;
	struct FileInfoBlock *fib;
	struct verzeichnis *naechstes,*letztes;

	if (verbose) Printf ("lese Verzeichnis: %s\n",mutter->name);

	naechstes = (struct verzeichnis *) verzeichnisse.mlh_TailPred; /* letzten Eintrag der Liste merken */

	if (fib = AllocDosObject (DOS_FIB,NULL))
		{
		if (lock = Lock (mutter->name,SHARED_LOCK))
			{
			if (Examine(lock,fib))
				{
				if (fib->fib_DirEntryType > 0)
					{
					while (!abbruch_pruefen() && ExNext(lock,fib))
						{
						if (fib->fib_DirEntryType > 0)
							neues_verzeichnis (mutter,fib->fib_FileName);
						else
							neue_datei (mutter,fib->fib_FileName,fib->fib_Size);
						}
					}
				}
			UnLock (lock);
			}
		FreeDosObject (DOS_FIB,fib);
		}

/*

   Unterverzeichnisse lesen: Wir lesen nur die Verzeichnisse, die wir eingefgt haben.
   Dazu haben wir uns am Anfang den letzten Eintrag in der Liste in "naechstes" gemerkt.
   Alle neuen Eintrge wurden danach eingefgt. Jetzt merken wir uns noch den Eintrag,
   der jetzt der letzte ist in "letztes". Dies ist der letzte Eintrag, den wir eingefgt
   haben. Alle Unterverzeichnisse werden dahinter eingefgt und mssen von uns nicht
   mehr gelesen werden.

*/
	letztes = (struct verzeichnis *) verzeichnisse.mlh_TailPred;
	while (!abbruch_pruefen() && naechstes != letztes)
		{
		naechstes = (struct verzeichnis *) naechstes->node.mln_Succ;
		lies_verzeichnis (naechstes);
		}
	}

/*----------------------------------------------------------------------------*/
/* Die Liste der Dateien drucken                                              */
/*----------------------------------------------------------------------------*/

	void drucke_dateien (void)

	{
	struct datei *datei, *naechste;
	char name[1000];

	for (datei = (struct datei *) dateien.mlh_Head; !abbruch_pruefen() && (naechste = (struct datei *) datei->node.mln_Succ); datei = naechste)
		{
		strcpy (name,datei->mutter->name);
		AddPart (name,datei->name,1000);
		Printf ("   %8ld %s\n",datei->groesse,name);
		}
	}

/*----------------------------------------------------------------------------*/
/* Alle Nodes einer Liste freigeben                                           */
/*----------------------------------------------------------------------------*/

	void liste_freigeben (struct MinList *list)

	{
	struct Node *node;

	while (node = RemHead ((struct List *)list))
		free (node);
	}

/*----------------------------------------------------------------------------*/
/* Hauptprogramm                                                              */
/*----------------------------------------------------------------------------*/

	int main (void)

	{
	struct RDArgs *rdargs;
	struct {
		char **path;
		long verbose;
		} args = {0};
	char **p;

	rdargs = ReadArgs ("PATH/A/M,VERBOSE/S",(LONG *)&args,NULL);
	if (!rdargs)
		{
		PrintFault (IoErr(),"Fehler in Argumenten");
		return (RETURN_ERROR);
		}

	if (args.verbose)
		verbose = TRUE;

	NewList ((struct List *)&dateien);
	NewList ((struct List *)&verzeichnisse);

	for (p = args.path; *p; p++)
		{
		neues_verzeichnis (NULL,*p);
		lies_verzeichnis ((struct verzeichnis *)verzeichnisse.mlh_TailPred);
		}

	drucke_dateien();

	liste_freigeben (&dateien);
	liste_freigeben (&verzeichnisse);

	FreeArgs (rdargs);

	return (RETURN_OK);
	}

/*----------------------------------------------------------------------------*/
/* Ende des Quelltextes                                                       */
/*----------------------------------------------------------------------------*/
